from building_blocks.models.querysets import BulkUpdateCreateQuerySet
from django.db import models, transaction
from django.db.models import Q

from .enums import Ordering
from .utils.itertools import limit_iterator


class RemoteResourceQuerySet(BulkUpdateCreateQuerySet, models.QuerySet):
    """
    Represents a QuerySet for an object that is a remote resource.
    Generally used alongside models who inherit from RemoteResource

    ...

    Methods
    -------
    download(max_pages=None, *args, **kwargs)
        Download and save remote objects to the DB from the API
    """

    def _bulk_update_or_create_helper(self, obj_list):
        model = self.model

        if model.remote_data_key_field is None:
            key_field = 'pk'
        else:
            key_field = model.remote_to_model_fields_map[model.remote_data_key_field]

        return self.bulk_update_or_create(obj_list, key_field, [
            field
            for field in model.remote_to_model_fields_map.values()
            if field != key_field
        ])

    def _get_list_api_generator(self):
        """
        Example: return APIClient().list_users
        """
        return lambda *args, **kwargs: iter(())

    def _get_list_api_iterator(self, *args, **kwargs):
        list_api_generator = self._get_list_api_generator()
        return list_api_generator(*args, **kwargs)

    def _save_objects_from_iterator(self, iterator):
        for data_list in iterator:
            with transaction.atomic():
                yield self._bulk_update_or_create_helper([
                    self.model.from_remote_data(item)
                    for item in data_list
                ])

    def download(self, max_pages=None, *args, **kwargs):
        iterator = limit_iterator(self._get_list_api_iterator(*args, **kwargs), max_pages)
        return self._save_objects_from_iterator(iterator)


class TimeSeriesQuerySetMixin(models.QuerySet):
    get_latest_by_field = None
    filter_q = None

    def _get_get_latest_by_field(self):
        return self.get_latest_by_field or self.model._meta.get_latest_by

    def _get_useful_timerange_qs(self):
        return self.filter(self.filter_q or Q())

    def _get_earliest_dt(self):
        get_latest_by_field = self._get_get_latest_by_field()
        filtered_qs = self._get_useful_timerange_qs()
        if obj := filtered_qs.order_by(get_latest_by_field).first():
            return getattr(obj, get_latest_by_field)

    def _get_latest_dt(self):
        return self.reverse()._get_earliest_dt()


class AscTimeSeriesRemoteResourceQuerySet(TimeSeriesQuerySetMixin, RemoteResourceQuerySet):
    def _get_list_api_iterator(self, refresh=False, *args, **kwargs):
        if refresh:
            start_dt, end_dt = None, None
        else:
            start_dt, end_dt = self._get_latest_dt(), None

        return super(AscTimeSeriesRemoteResourceQuerySet, self)._get_list_api_iterator(
            ordering=Ordering.earlier_first,
            start_dt=start_dt,
            end_dt=end_dt,
            *args, **kwargs,
        )


class DescTimeSeriesRemoteResourceQuerySet(TimeSeriesQuerySetMixin, RemoteResourceQuerySet):
    def _get_list_api_iterator(self, fill=False, refresh=False, *args, **kwargs):
        if refresh:
            start_dt, end_dt = None, None
        elif fill:
            start_dt, end_dt = None, self._get_earliest_dt()
        else:
            start_dt, end_dt = self._get_latest_dt(), None

        return super(DescTimeSeriesRemoteResourceQuerySet, self)._get_list_api_iterator(
            ordering=Ordering.later_first,
            start_dt=start_dt,
            end_dt=end_dt,
            *args, **kwargs,
        )



import json

from django.db import models

from .querysets import RemoteResourceQuerySet


def _sanitize_json_for_db(item):
    return json.loads(json.dumps(item).replace('\\u0000', ''))


class RemoteResource(
    models.Model
):
    class Meta:
        abstract = True

    remote_to_model_fields_map = {}
    remote_data_key_field = None

    _json = models.JSONField()

    objects = RemoteResourceQuerySet.as_manager()

    @classmethod
    def from_remote_data(cls, item, **kwargs):
        fields_dict = {
            model_field: item[remote_field]
            for remote_field, model_field in cls.remote_to_model_fields_map.items()
            if remote_field in item
        }
        fields_dict.update(kwargs)
        return cls(
            _json=_sanitize_json_for_db(item),
            **fields_dict,
        )



from abc import ABC
from typing import Tuple, Any

from django.db import models
from django.db.models import F


class HasCachedPropertiesQuerySet(ABC, models.QuerySet):
    """
    Represents a QuerySet for an object that has cached properties.
    Generally used alongside models who inherit from HasCachedProperties

    You'd want to override and define method _refresh_annotations

    ...

    Attributes
    ----------
    cached_properties : list
        a list of properties that are cached

    Methods
    -------
    ready(fresh=False)
        Annotate all property fields. Load the values from cache or freshly calculate them.

    refresh()
        Flush the cache, refresh the annotations and then update the cache with the new values
    """

    cached_properties = []

    def _flush_cache(self):
        return self.update(**{
            f'_cached_{field}': None
            for field in self.cached_properties
        })

    def update_cache(self, *args):
        return self.update(**{
            f'_cached_{field}': F(field)
            for field in (args or self.cached_properties)
        })

    def _load_cache(self):
        return self.annotate(**{
            field: F(f'_cached_{field}')
            for field in self.cached_properties
        })

    def _refresh_annotations(self) -> 'HasCachedPropertiesQuerySet':
        """
        Override this method for the logic that loads up the values for the properties as queryset annotations.

        Returns
        -------
        HasCachedPropertiesQuerySet
            a queryset with all property fields annotated from fresh values
        """
        raise NotImplementedError

    def ready(self, fresh=False) -> 'HasCachedPropertiesQuerySet':
        """
        Annotates all property fields. Loads the values from cache if fresh is false or freshly calculates them if true.

        Parameters
        ----------
        fresh : bool, optional
            A flag used to determine whether to calculate fresh data (if True) or load from cache (if False, default)

        Returns
        -------
        HasCachedPropertiesQuerySet
            a queryset with all property fields annotated from fresh values or values loaded from the cache
        """
        return self._refresh_annotations() if fresh else self._load_cache()

    def refresh(self, flush_cache=True) -> tuple['HasCachedPropertiesQuerySet', Any]:
        """
        Flush the cache, refresh the annotations and then update the cache with the new values

        Returns
        -------
        HasCachedPropertiesQuerySet
            A queryset with all property fields annotated from fresh values
        """
        if flush_cache:
            self._flush_cache()
        annotated_qs = self._refresh_annotations()
        return annotated_qs, annotated_qs.update_cache()



from django.core.management.base import BaseCommand

from .mixins import RefreshableCommandMixin, FillableCommandMixin


class DownloadResourceCommand(BaseCommand):
    model = None
    queryset = None
    accumulate_qs = False
    download_options = ('max_pages',)

    @property
    def _opts(self):
        return self.model._meta

    @property
    def help(self):
        return f"Download {self._opts.verbose_name_plural} into the database from a remote resource"

    def _write_success(self, message):
        self.stdout.write(self.style.SUCCESS(message))

    def _write_success_page_downloaded(self, qs, page):
        count = qs.count()
        self._write_success(f"({page}) Created or updated {count} {self._opts.verbose_name_plural}")

    def _write_success_done(self, *args, **kwargs):
        total_count = kwargs['total_count']
        self._write_success(f"Created or updated {total_count} {self._opts.verbose_name_plural} in total")

    def _write_success_page_post_processed(self, results, page):
        pass

    def get_queryset(self):
        if self.queryset is None:
            return self.model.objects.all()
        return self.queryset

    def _pick_options(self, options):
        return {k: v for k, v in options.items() if k in self.download_options}

    def download(self, **kwargs):
        return self.get_queryset().download(**kwargs)

    def post_process_page(self, qs):
        return qs

    def post_process_all(self, accum_qs):
        return accum_qs

    def add_arguments(self, parser):
        parser.add_argument('--maxpages', type=int, dest='max_pages')

    def handle(self, *args, **options):
        total_count = 0
        accum_qs = self.get_queryset().none()

        for page, qs in enumerate(self.download(**self._pick_options(options))):
            self._write_success_page_downloaded(qs, page)
            total_count += qs.count()
            if self.accumulate_qs:
                accum_qs |= qs
            post_process_qs = self.post_process_page(qs)
            self._write_success_page_post_processed(post_process_qs, page)

        self.post_process_all(accum_qs)
        self._write_success_done(total_count=total_count)


class DownloadTimeSeriesResourceCommand(RefreshableCommandMixin, DownloadResourceCommand):
    download_options = (
        *DownloadResourceCommand.download_options,
        *RefreshableCommandMixin.download_options,
    )


class DownloadAscTimeSeriesResourceCommand(DownloadTimeSeriesResourceCommand):
    pass


class DownloadDescTimeSeriesResourceCommand(
    RefreshableCommandMixin,
    FillableCommandMixin,
    DownloadResourceCommand
):
    download_options = (
        *DownloadResourceCommand.download_options,
        *RefreshableCommandMixin.download_options,
        *FillableCommandMixin.download_options,
    )


__all__ = [
    'DownloadResourceCommand',
    'DownloadTimeSeriesResourceCommand',
    'DownloadAscTimeSeriesResourceCommand',
    'DownloadDescTimeSeriesResourceCommand'
]



from django.core.paginator import Paginator
from django.db import models
from django.db.models import Min, Max


class PageableQuerySet(models.QuerySet):
    """

    """

    def paginate(self, limit):
        d = self.aggregate(min=Min('id'), max=Max('id'))
        min_id, max_id = d['min'], d['max']
        if min_id is None:
            return self
        for i in range(min_id, max_id + 1, limit):
            yield self.filter(id__gte=i, id__lt=i + limit)

    def paginate_dprc(self, limit, simple=True, mutating=False):
        """

        :param limit:
        :param simple:
        :param mutating:
        :return:
        """

        qs = self.model.objects.all() if simple else self
        pk_values = self.values_list('pk', flat=True)

        if not mutating:
            for page in Paginator(pk_values, limit):
                yield qs.filter(pk__in=page.object_list)
        else:
            paginated_pks = tuple(
                tuple(page.object_list)
                for page in Paginator(pk_values, limit)
            )
            while paginated_pks:
                pks_page = paginated_pks[0]
                yield qs.filter(pk__in=pks_page)
                paginated_pks = paginated_pks[1:]



class RefreshableCommandMixin:
    download_options = (
        'refresh',
    )

    def add_arguments(self, parser):
        super(RefreshableCommandMixin, self).add_arguments(parser)
        parser.add_argument('--refresh', action='store_true')


class FillableCommandMixin:
    download_options = (
        'fill',
    )

    def add_arguments(self, parser):
        super(FillableCommandMixin, self).add_arguments(parser)
        parser.add_argument('--fill', action='store_true')


__all__ = [
    'RefreshableCommandMixin',
    'FillableCommandMixin',
]
